@@ -45,9 +45,9 @@
 //!    let next = format!("{}/?limit=1&offset=6&a__lt=10&ordering=-a", server.uri());
 //!    assert_eq!(body, serde_json::json!{
 //!      {
-//!        "count": 1,
+//!        "count": 10,
 //!        "next": next,
-//!        "prev": prev,
+//!        "previous": prev,
 //!        "results": [
 //!          { "a": 4 }
 //!        ]
@@ -89,14 +89,21 @@ pub enum MockError {
 
 struct ResponseSet<T> {
     contents: Vec<T>,
+    total_matches: usize,
     next: Option<String>,
     prev: Option<String>,
 }
 
 impl<T> ResponseSet<T> {
-    pub fn new(contents: Vec<T>, next: Option<String>, prev: Option<String>) -> Self {
+    pub fn new(
+        contents: Vec<T>,
+        total_matches: usize,
+        next: Option<String>,
+        prev: Option<String>,
+    ) -> Self {
         ResponseSet {
             contents,
+            total_matches,
             next,
             prev,
         }
@@ -108,7 +115,7 @@ impl<T: IntoRow> ResponseSet<&T> {
         let mut map = serde_json::map::Map::new();
         map.insert(
             "count".to_string(),
-            serde_json::Value::Number(serde_json::Number::from(self.contents.len())),
+            serde_json::Value::Number(serde_json::Number::from(self.total_matches)),
         );
         map.insert("results".to_string(), to_rows(&self.contents));
         map.insert(
@@ -119,7 +126,7 @@ impl<T: IntoRow> ResponseSet<&T> {
                 .unwrap_or(serde_json::Value::Null),
         );
         map.insert(
-            "prev".to_string(),
+            "previous".to_string(),
             self.prev
                 .as_ref()
                 .map(|x| serde_json::Value::String(x.clone()))
@@ -145,6 +152,7 @@ struct Page {
 
 struct PaginatedResponse<'a, T> {
     data: Vec<&'a T>,
+    total: usize,
     next: Option<Page>,
     prev: Option<Page>,
 }
@@ -158,16 +166,16 @@ struct ResponseSetBuilder<T> {
 
 impl<T> Default for ResponseSetBuilder<T> {
     fn default() -> Self {
-        Self::new()
+        Self::new(None)
     }
 }
 
 impl<T> ResponseSetBuilder<T> {
-    pub fn new() -> Self {
+    pub fn new(default_limit: Option<usize>) -> Self {
         ResponseSetBuilder {
             ordering: Vec::new(),
             filtering: Vec::new(),
-            limit: None,
+            limit: default_limit,
             offset: 0,
         }
     }
@@ -207,6 +215,7 @@ impl<T> ResponseSetBuilder<T> {
         }
 
         let limit = self.limit.unwrap_or(v.len());
+        let total = v.len();
 
         let start = min(v.len(), self.offset);
         let prev = if start > 0 {
@@ -234,6 +243,7 @@ impl<T> ResponseSetBuilder<T> {
 
         PaginatedResponse {
             data: v,
+            total,
             next,
             prev,
         }
@@ -246,6 +256,8 @@ fn make_page_url(url: &Url, offset: usize, limit: usize) -> Url {
     let mut new_pairs = new_url.query_pairs_mut();
     let old_pairs = url.query_pairs();
 
+    let mut offset_seen = false;
+
     new_pairs.clear();
     for (key, value) in old_pairs {
         match key.as_ref() {
@@ -256,12 +268,18 @@ fn make_page_url(url: &Url, offset: usize, limit: usize) -> Url {
                 if offset > 0 {
                     new_pairs.append_pair("offset", &offset.to_string());
                 }
+                offset_seen = true;
             }
             _ => {
                 new_pairs.append_pair(key.as_ref(), value.as_ref());
             }
         }
     }
+
+    if !offset_seen && offset > 0 {
+        new_pairs.append_pair("offset", &offset.to_string());
+    }
+
     new_pairs.finish();
     drop(new_pairs);
 
@@ -319,8 +337,10 @@ impl<T> RowSource for std::sync::Arc<Vec<T>> {
 }
 
 fn parse_query<'a, T, R>(
-    url: &Url,
+    input_url: &Url,
+    output_url: &Url,
     iter: <&'a <<T as RowSource>::Rows as Deref>::Target as IntoIterator>::IntoIter,
+    default_limit: Option<usize>,
 ) -> Result<ResponseSet<&'a R>, MockError>
 where
     T: Send + Sync + RowSource<Item = R>,
@@ -328,10 +348,10 @@ where
     <T as RowSource>::Rows: Deref,
     for<'t> &'t <<T as RowSource>::Rows as Deref>::Target: IntoIterator<Item = &'t R>,
 {
-    let mut rb = ResponseSetBuilder::new();
+    let mut rb = ResponseSetBuilder::new(default_limit);
     let qr = OperatorSet::<R>::new();
     let sr = OrderingSet::<R>::new();
-    let pairs = url.query_pairs();
+    let pairs = input_url.query_pairs();
     for (key, value) in pairs {
         match key.as_ref() {
             "ordering" => {
@@ -358,12 +378,13 @@ where
     let response = rb.apply(iter);
     Ok(ResponseSet::new(
         response.data,
+        response.total,
         response
             .next
-            .map(|page| make_page_url(url, page.offset, page.limit).to_string()),
+            .map(|page| make_page_url(output_url, page.offset, page.limit).to_string()),
         response
             .prev
-            .map(|page| make_page_url(url, page.offset, page.limit).to_string()),
+            .map(|page| make_page_url(output_url, page.offset, page.limit).to_string()),
     ))
 }
 
@@ -383,6 +404,7 @@ where
 pub struct Endpoint<T> {
     row_source: T,
     base_uri: Option<Url>,
+    default_limit: Option<usize>,
 }
 
 impl<T, R> Endpoint<T>
@@ -409,8 +431,20 @@ where
         Self {
             row_source,
             base_uri: base_uri.map(|x| Url::parse(x).unwrap()),
+            default_limit: None,
         }
     }
+
+    /// Set the default number of results returned
+    ///
+    /// This is the value used when no limit is specified in the
+    /// query.  This value is configurable in Django; the default
+    /// behaviour of this mock endpoint is to return everything, but
+    /// that makes testing code that uses the default pagination more
+    /// difficult.
+    pub fn default_limit(&mut self, limit: usize) {
+        self.default_limit = Some(limit);
+    }
 }
 
 impl<T, R> Respond for Endpoint<T>
@@ -430,7 +464,7 @@ where
             u.set_scheme(base.scheme()).unwrap();
             u.set_port(base.port()).unwrap();
         }
-        let body = parse_query::<T, R>(&u, (&data).into_iter());
+        let body = parse_query::<T, R>(&request.url, &u, (&data).into_iter(), self.default_limit);
         match body {
             Ok(rs) => ResponseTemplate::new(200).set_body_json(rs.mock_json()),
             Err(e) => {
@@ -441,44 +475,25 @@ where
     }
 }
 
+/// Match a Django-style nested endpoint in wiremock
+///
+/// This avoids using regular expressions in the calling code. Here
+/// - `root` is the base of Django API, like `"/api/v0.2"`
+/// - `parent` is the parent object, like `"bar"`
+/// - `child` is the nested object, like `"foo"`
+///
+/// See the description of [NestedEndpointParams] for more detail.
+pub fn nested_endpoint_matches(root: &str, parent: &str, child: &str) -> impl wiremock::Match {
+    wiremock::matchers::path_regex(format!(r"^{}/{}/[^/]+/{}/$", root, parent, child))
+}
+
 fn replace_into(target: &str, cap: &Captures<'_>) -> String {
     let mut res = String::new();
     cap.expand(target, &mut res);
     res
 }
 
-#[allow(rustdoc::bare_urls)]
-/// Use a regex to match a query path, mutating both the path and query string.
-///
-/// Some Django-based services support nested endpoints, where the
-/// path of the URL might look like:
-///
-///    <http://example.com/womble/6/neighbours>
-///
-/// Since our parsing operates on the query parameters, we'd like to rewrite this
-/// URL before we process it, to fit the remainder of the pipeline better. For
-/// example we might want to create:
-///
-///    <http://example.com/neighbours?womble=6>
-///
-/// from the preceding URL.
-///
-/// Example:
-/// ```rust
-/// use wiremock::http::Url;
-/// use django_query::mock::UrlTransform;
-///
-/// let ut = UrlTransform::new(r"^/base/foos/(\d+)/(.+)$",
-///                            r"/base/$2",
-///                            vec![("foo", "$1")]);
-/// let u = Url::parse("http://example.com/base/foos/1214/bars?ordering=name").unwrap();
-/// let t = Url::parse("http://example.com/base/bars?ordering=name&foo=1214").unwrap();
-/// assert_eq!(ut.transform(&u), t);
-/// ```
-///
-/// This is intended for use with [NestedEndpoint], which differs from
-/// [Endpoint] only in applying a [UrlTransform] to its input.
-pub struct UrlTransform {
+struct UrlTransform {
     regex: Regex,
     path: String,
     pairs: Vec<(String, String)>,
@@ -497,6 +512,7 @@ impl UrlTransform {
     }
 
     pub fn transform(&self, url: &Url) -> Url {
+        debug!("Matching {} against {:?}", url, self.regex);
         if let Some(captures) = self.regex.captures(url.path()) {
             let mut u = url.clone();
             u.set_path(&replace_into(&self.path, &captures));
@@ -513,11 +529,13 @@ impl UrlTransform {
 
 /// A Django nested route.
 ///
-/// A nested route in Django is a way of specifying that a particular filter field must
-/// be present when querying for a particular object type, and that it the filter type
-/// must be exact match. So for example, if all objects of type `Foo` have a reference to
-/// a `Bar`, and its too expensive or undesirable to query for `Foo`s without specifying
-/// which `Bar` they come from, then you can express this in Django as a nested route:
+/// A nested route in Django is a way of specifying that a particular
+/// filter field must be present when querying for a particular object
+/// type, and that the filter type must be exact match. So for
+/// example, if all objects of type `Foo` have a reference to a `Bar`,
+/// and its too expensive or undesirable to query for `Foo`s without
+/// specifying which `Bar` they come from, then you can express this
+/// in Django as a nested route:
 ///
 ///   ``http://example.com/bars/my-bar-id/foos``
 ///
@@ -525,14 +543,36 @@ impl UrlTransform {
 /// identify a particular associated `Bar`.
 ///
 /// Since the django-query crate operates on query parameters, the
-/// simplest way to handle this is to use a [UrlTransform] to convert
-/// any relevant parts of the path into query parameters as required,
-/// which is all this type does.  It is otherwise identical to
-/// [Endpoint].
+/// simplest way to handle this is to convert any relevant parts of
+/// the path into query parameters as required, which is all this type
+/// does. See its construction parameters [NestedEndpointParams] for
+/// the details. This type is otherwise identical to [Endpoint].
 pub struct NestedEndpoint<T> {
     transform: UrlTransform,
     row_source: T,
     base_uri: Option<Url>,
+    default_limit: Option<usize>,
+}
+
+/// The construction parameters for [NestedEndpoint]
+pub struct NestedEndpointParams<'a, T> {
+    /// The base mock path, without trailing slash, for example `"/api/v0.2"`
+    pub root: &'a str,
+    /// The objects this endpoint is nested under, for example `"bars"`.
+    pub parent: &'a str,
+    /// The objects this endpoint serves, for example `"foos"`
+    pub child: &'a str,
+    /// The filter path for the parent from the child, so for
+    /// example `"parent"` if the filter expression for a given
+    /// Foo to obtain its containing Bar is `"parent"`.
+    pub parent_query: &'a str,
+    /// The actual data to serve, as for [Endpoint]
+    pub row_source: T,
+    /// The mock server URI, since wiremock does not make this
+    /// available to us; specifying None here will break
+    /// pagination (if requested) until wiremock propagates this
+    /// information.
+    pub base_uri: Option<&'a str>,
 }
 
 impl<T, R> NestedEndpoint<T>
@@ -544,16 +584,31 @@ where
 {
     /// Create a new NestedEndpoint.
     ///
-    /// The given [UrlTransform] will be applied to every incoming
-    /// request. The `row_source` will be used to gather the data to
-    /// respond to the rewritten request.
-    pub fn new(transform: UrlTransform, row_source: T, base_uri: Option<&str>) -> Self {
+    /// The arguments to this function are wrapped in their own
+    /// structure: [NestedEndpointParams].
+    pub fn new(p: NestedEndpointParams<'_, T>) -> Self {
         Self {
-            transform,
-            row_source,
-            base_uri: base_uri.map(|x| Url::parse(x).unwrap()),
+            transform: UrlTransform::new(
+                &format!(r"^{}/{}/(?P<parent>[^/]+)/{}/$", p.root, p.parent, p.child),
+                &format!("{}/{}/", p.root, p.child),
+                vec![(&*p.parent_query, "${parent}")],
+            ),
+            row_source: p.row_source,
+            base_uri: p.base_uri.map(|x| Url::parse(x).unwrap()),
+            default_limit: None,
         }
     }
+
+    /// Set the default number of results returned
+    ///
+    /// This is the value used when no limit is specified in the
+    /// query.  This value is configurable in Django; the default
+    /// behaviour of this mock endpoint is to return everything, but
+    /// that makes testing code that uses the default pagination more
+    /// difficult.
+    pub fn default_limit(&mut self, limit: usize) {
+        self.default_limit = Some(limit);
+    }
 }
 
 impl<T, R> Respond for NestedEndpoint<T>
@@ -565,16 +620,22 @@ where
 {
     fn respond(&self, request: &Request) -> ResponseTemplate {
         trace!("Request URL: {}", request.url);
-        let mut u = self.transform.transform(&request.url);
-        trace!("Transformed URL: {}", u);
+        let input_url = self.transform.transform(&request.url);
+        trace!("Transformed URL: {}", input_url);
 
         let data: <T as RowSource>::Rows = self.row_source.get();
+        let mut output_url = request.url.clone();
         if let Some(ref base) = self.base_uri {
-            u.set_host(base.host_str()).unwrap();
-            u.set_scheme(base.scheme()).unwrap();
-            u.set_port(base.port()).unwrap();
+            output_url.set_host(base.host_str()).unwrap();
+            output_url.set_scheme(base.scheme()).unwrap();
+            output_url.set_port(base.port()).unwrap();
         }
-        let body = parse_query::<T, R>(&u, (&data).into_iter());
+        let body = parse_query::<T, R>(
+            &input_url,
+            &output_url,
+            (&data).into_iter(),
+            self.default_limit,
+        );
         match body {
             Ok(rs) => ResponseTemplate::new(200).set_body_json(rs.mock_json()),
             Err(e) => {
